/*
 * FTM 
 * Copyright (C) 1994, 1995, 1998, 1999, 2007 by IRCAM-Centre Georges Pompidou, Paris, France.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.
 * 
 * See file COPYING.LIB for further informations on licensing terms.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */

#ifndef _SEQUENCE_H_
#define _SEQUENCE_H_

#include "fts.h"
#include <math.h>

#ifdef WIN32 
#define fmax(a,b) (((a)>(b))? (a) : (b))
#endif

/** 
 * @defgroup sequence sequence: sequence of time-tagged values
 * @ingroup  fts_classes
 *  
 *  sequence of time-tagged values
 */

/** 
 * @var fts_class_t *sequence_class
 * @brief sequence_class ... 
 * @ingroup sequence 
 */
FTS_API fts_class_t *sequence_class;

/**
 * @struct _sequence_
 * @brief sequence struct
 * @ingroup sequence
 */
struct _sequence_
{ 
  fts_object_t o; /**< fts_object ... */

  fts_class_t *type; /**< type of events */
  event_t *first; /**< pointer to first event */
  event_t *last; /**< pointer to last event */
  int size; /**< # of events in sequence */
  
  int editor_open;
};

#define sequence_set_type(s, t) ((s)->type = (t))
#define sequence_get_type(s) ((s)->type)

#define sequence_set_editor_open(t) ((t)->editor_open = 1)
#define sequence_set_editor_close(t) ((t)->editor_open = 0)
#define sequence_editor_is_open(t) ((t)->editor_open != 0)

#ifdef AVOID_MACROS

/** 
 * @fn event_t *sequence_get_first(sequence_t *sequence)
 * @brief get first event in sequence
 * @param sequence the sequence
 * @return first event in sequence
 * @ingroup sequence 
 */
event_t *sequence_get_first(sequence_t *sequence);

/** 
 * @fn event_t *sequence_get_last(sequence_t *sequence)
 * @brief get last event in sequence
 * @param sequence the sequence
 * @return last event in sequence
 * @ingroup sequence 
 */
event_t *sequence_get_last(sequence_t *sequence);

/** 
 * @fn int sequence_get_size(sequence_t *sequence)
 * @brief get sequence size
 * @param sequence the sequence
 * @return sequence size
 * @ingroup sequence 
 */
int sequence_get_size(sequence_t *sequence);

/** 
 * @fn double sequence_get_last_time(sequence_t *sequence)
 * @brief get timetag of last event in sequence
 * @param sequence the sequence
 * @return timetag of last event
 * @ingroup sequence 
 */
double sequence_get_last_time(sequence_t *sequence);

/** 
 * @fn double sequence_get_duration(sequence_t *sequence)
 * @brief get sequence duration
 * @param sequence the sequence
 * @return sequence duration
 * @ingroup sequence 
 */
double sequence_get_duration(sequence_t *sequence);

#else

#define sequence_get_first(s) ((s)->first)
#define sequence_get_last(s) ((s)->last)
#define sequence_get_size(s) ((s)->size)
#define sequence_get_last_time(s) (((s)->last)? (s)->last->time: 0.0)

#define sequence_get_duration(s) (sequence_get_last_time(s))

#endif

FTS_API double sequence_get_total_duration(sequence_t *sequence);

/** 
 * @fn void sequence_add_event(sequence_t *sequence, double time, event_t *event)
 * @brief add event in sequence at given time
 * @param sequence the sequence
 * @param time timetag for event
 * @param event to be added event
 * @ingroup sequence 
 */
FTS_API void sequence_add_event(sequence_t *sequence, double time, event_t *event);

/** 
 * @fn void sequence_add_event_after(sequence_t *sequence, double time, event_t *event, event_t *after)
 * @brief add event in sequence at given time after given event
 * @param sequence the sequence
 * @param time timetag for event
 * @param event to be added event
 * @param after add new event after this one ...
 * @ingroup sequence 
 */
FTS_API void sequence_add_event_after(sequence_t *sequence, double time, event_t *event, event_t *after);

/** 
* @fn void sequence_add_event_before(sequence_t *sequence, double time, event_t *event, event_t *before)
* @brief add event in sequence at given time before given event
* @param sequence the sequence
* @param time timetag for event
* @param event to be added event
* @param after add new event before this one ...
* @ingroup sequence 
*/
FTS_API void sequence_add_event_before(sequence_t *sequence, double time, event_t *event, event_t *before);

/** 
 * @fn void sequence_append_event(sequence_t *sequence, double time, event_t *event)
 * @brief append event in sequence
 * @param sequence the sequence
 * @param time timetag for event
 * @param event to be added event
 * @ingroup sequence 
 */
FTS_API void sequence_append_event(sequence_t *sequence, double time, event_t *event);

/** 
 * @fn void sequence_move_event(sequence_t *sequence, event_t *event, double time)
 * @brief move event in sequence
 * @param sequence the sequence
 * @param event to be added event
 * @param time new timetag for event
 * @ingroup sequence 
 */
FTS_API void sequence_move_event(sequence_t *sequence, event_t *event, double time);

/** 
 * @fn void sequence_remove_event(sequence_t *sequence, event_t *event)
 * @brief remove event from sequence
 * @param sequence the sequence
 * @param event to be removed event
 * @ingroup sequence 
 */
FTS_API void sequence_remove_event(sequence_t *sequence, event_t *event);

/** 
 * @fn void sequence_copy(sequence_t *org, sequence_t *copy)
 * @brief copy sequence content in given sequence
 * @param org the original sequence
 * @param copy copy sequence
 * @ingroup sequence 
 */
FTS_API void sequence_copy(sequence_t *org, sequence_t *copy);

/** 
 * @fn void sequence_merge(sequence_t *sequence, sequence_t *merge)
 * @brief merge sequence content whit/in given sequence
 * @param sequence first sequence
 * @param merge merge in sequence
 * @ingroup sequence 
 */
FTS_API void sequence_merge(sequence_t *sequence, sequence_t *merge);

/** 
 * @fn void sequence_clear(sequence_t *sequence)
 * @brief remove all events from sequence and associated markers sequence
 * @param sequence the sequence
 * @ingroup sequence 
 */
FTS_API void sequence_clear(sequence_t *sequence);

/**
 * @brief prepare iterator over all event's values, possibly in range given by varargs
 */
FTS_API void sequence_get_values(sequence_t *sequence, fts_iterator_t *iter, int ac, const fts_atom_t *at);

/**
 * @brief prepare iterator over all events, possibly in range given by varargs
 */
FTS_API void sequence_get_events(sequence_t *sequence, fts_iterator_t *iter, int ac, const fts_atom_t *at);

/**
 * @brief prepare iterator over all events of several sequences merged
 */
FTS_API void sequence_get_events_merged (fts_iterator_t *iter, int n, sequence_t *seqs[]);

/**
 * @brief return sequence index of current event of merge iterator (call after "next")
 */
FTS_API int sequence_merge_iterator_get_index (fts_iterator_t *iter);

/** 
 * @fn event_t *sequence_get_event_by_time(sequence_t *sequence, double time)
 * @brief get first event at or after given time
 * @param sequence the sequence
 * @param time the time
 * @return the event
 * @ingroup sequence 
 */
FTS_API event_t *sequence_get_event_by_time(sequence_t *sequence, double time);


/** 
* @fn event_t *sequence_get_event_by_index(sequence_t *sequence, int index)
* @brief get event at given index
* @param sequence the sequence
* @param index the index
* @return the event
* @ingroup sequence 
*/
FTS_API event_t *sequence_get_event_by_index(sequence_t *sequence, int index);

/** 
 * @fn event_t *sequence_get_next_by_time(sequence_t *sequence, double time)
 * @brief get next event after given time
 * @param sequence the sequence
 * @param time the time
 * @return the event
 * @ingroup sequence 
 */
FTS_API event_t *sequence_get_next_by_time(sequence_t *sequence, double time);

/** 
 * @fn event_t *sequence_get_next_by_time_after(sequence_t *sequence, double time, event_t *here)
 * @brief get next event after given time. The search is started from given element.
 * @param sequence the sequence
 * @param time the time
 * @param here search start from this event
 * @return the event
 * @ingroup sequence 
 */
FTS_API event_t *sequence_get_next_by_time_after(sequence_t *sequence, double time, event_t *here);

/** 
 * @fn  event_t *sequence_get_left_by_time_from(sequence_t *sequence, double time, event_t *here)
 * @brief get event before or at given time. Search in descending time order starts at given event
 * @param sequence the sequence
 * @param time the time
 * @param here search starts from this event
 * @return the event
 * @ingroup sequence 
 * 
 * get event before or at <time>, search in descending time order starts at event <here>, if not NULL 
 * (N.B.: if time is before first event, the first event is returned(?)
 */
FTS_API event_t *sequence_get_left_by_time_from(sequence_t *sequence, double time, event_t *here);

/** 
 * @fn void sequence_segment_get(sequence_t *sequence, double begin, double end, event_t **first, event_t **after)
 * @brief  get segment by time
 * @param sequence the sequence
 * @param begin starting time
 * @param end end time
 * @param first start search from this event
 * @param after next event after the segment
 * @ingroup sequence 
 */
FTS_API void sequence_segment_get(sequence_t *sequence, double begin, double end, event_t **first, event_t **after);

/** 
 * @fn void sequence_segment_shift(sequence_t *sequence, event_t *first, event_t *after, double begin, double end, double shift)
 * @brief move every event in segment of given time
 * @param sequence the sequence
 * @param first start search from this event
 * @param after next event after the segment
 * @param begin starting time
 * @param end end time
 * @param shift shift in milliseconds
 * @ingroup sequence 
 */
FTS_API void sequence_segment_shift(sequence_t *sequence, event_t *first, event_t *after, double begin, double end, double shift);

/** 
 * @fn void sequence_segment_stretch(sequence_t *sequence, event_t *first, event_t *after, double begin, double end, double stretch)
 * @brief stretch every event in segment of given stretch arg
 * @param sequence the sequence
 * @param first start search from this event
 * @param after next event after the segment
 * @param begin starting time
 * @param end end time
 * @param stretch ...
 * @ingroup sequence 
 */
FTS_API void sequence_segment_stretch(sequence_t *sequence, double begin, double end, double stretch);

/** 
 * @fn void sequence_segment_quantize(sequence_t *sequence, event_t *first, event_t *after, double begin, double end, double quantize)
 * @brief quantize segment ...
 * @param sequence the sequence
 * @param first start search from this event
 * @param after next event after the segment
 * @param begin starting time
 * @param end end time
 * @param quantize ...
 * @ingroup sequence 
 */
FTS_API void sequence_segment_quantize(sequence_t *sequence, double begin, double end, double quantize);

//void sequence_highlight_event(sequence_t *sequence, event_t *event);
//void sequence_highlight_events(sequence_t *sequence, int n, event_t *event[]);
//void sequence_highlight_time(sequence_t *sequence, double time);
//void sequence_highlight_events_and_time(sequence_t *sequence, double time, int n, event_t *event[]);
//void sequence_highlight_cluster(sequence_t *sequence, event_t *event, event_t *next);
//event_t *sequence_highlight_and_next(sequence_t *sequence, event_t *event);
//void sequence_highlight_reset(sequence_t *sequence);

FTS_API void seqclient_update_event(sequence_t *sequence, event_t *event);
FTS_API void seqclient_update(sequence_t *sequence);
FTS_API void seqclient_add_event(sequence_t *sequence, event_t *event);
FTS_API void seqclient_remove_event(sequence_t *sequence, event_t *event);
FTS_API void seqclient_move_event(sequence_t *sequence, event_t *event, double time);
FTS_API void seqclient_clear(sequence_t *sequence);

#endif
