/*
 * FTM 
 * Copyright (C) 1994, 1995, 1998, 1999, 2007 by IRCAM-Centre Georges Pompidou, Paris, France.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.
 * 
 * See file COPYING.LIB for further informations on licensing terms.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */

#ifndef _SEQUENCE_TRACK_H_
#define _SEQUENCE_TRACK_H_

#include "fts.h"
#include <math.h>

#ifdef WIN32 
#define fmax(a,b) (((a)>(b))? (a) : (b))
#endif

/** 
 * @defgroup track track: sequence of time-tagged values
 * @ingroup  fts_classes
 *  
 *  sequence of time-tagged values
 */

/** 
 * @var fts_class_t *track_class
 * @brief track_class ... 
 * @ingroup track 
 */
FTS_API fts_class_t *track_class;

/**
 * @struct _track_
 * @brief track struct
 * @ingroup track
 */
typedef struct _track_ track_t;

struct _track_
{ 
  fts_object_t o;
  fts_class_t *type;
  sequence_t *events;
  sequence_t *markers;
  sequence_t *tempo;
  sequence_t *metrics;
  
  fts_hashtable_t *more;

  struct _track_editor *editor; /**< track editor */
  int editor_open; /**< flag: is 1 if track editor is open */
  int save_editor; /**< flag: is 1 if want to save editor state */	
};

#define MARKERS_BAR_TOLERANCE 20.0 /* tolerance for bars */

#define track_get_type(t) ((t)->type)
#define track_get_size(t) ((t)->events->size)

#define track_get_duration(t) (track_get_markers(t)? fmax(sequence_get_last_time((t)->markers), sequence_get_last_time((t)->events)): sequence_get_last_time((t)->events))

#define track_do_save_editor(t) 1 /*((t)->save_editor != 0)*/
#define track_is_in_multitrack(t) 0

#define track_get_events(t) ((t)->events)
#define track_get_markers(t) ((t)->markers)
#define track_get_tempo(t) ((t)->tempo)
#define track_get_metrics(t) ((t)->metrics)

#define track_set_editor_open(t) ((t)->editor_open = 1)
#define track_set_editor_close(t) ((t)->editor_open = 0)
#define track_editor_is_open(t) ((t)->editor_open != 0)

FTS_API event_t *track_insert_value(track_t *track, double time, const fts_atom_t *a);
/**
 * @brief insert given marker into marker track at given time, or
 * create new marker object with given arguments */
FTS_API event_t *track_insert_marker(track_t *track, double time, int ac, const fts_atom_t *a);
FTS_API event_t *track_insert_tempo(track_t *track, double time, double tempo);
FTS_API event_t *track_insert_bar(track_t *track, double time, fts_symbol_t meter);
FTS_API void track_merge(track_t *track_merge, track_t *track_src);
FTS_API void track_renumber_bars(track_t *track);
FTS_API void track_upload(track_t *track);

FTS_API void  track_copy(track_t *org, track_t *copy);
FTS_API void track_clear(track_t *track);
FTS_API void track_clear_events(track_t *track);
FTS_API void track_clear_markers(track_t *track);
FTS_API void track_clear_tempo(track_t *track);
FTS_API void track_clear_metrics(track_t *track);


/** @name Functions for highlight events in track editor
*  @{ */

/**
* @fn void track_highlight_event(track_t *track, event_t *event)
 * @brief highlight event in track editor
 * @param track the track
 * @param event	pointer to track event
 * @ingroup track
 *
 * Highlight one event in track editor and set cursor on its starting time.
 * Previous highlighting is removed.
 */
FTS_API void track_highlight_event(track_t *track, event_t *event);

/**
* @fn void track_highlight_events(track_t *track, int n, event_t *event[])
 * @brief highlight events in track editor
 * @param track the track
 * @param n	number of events in event array
 * @param event	array of track events
 * @ingroup track
 *
 * Highlight events in track editor (up to 64) and set cursor to start time of group.
 * Previous highlighting is removed.
 */
FTS_API void track_highlight_events(track_t *track, int n, event_t *event[]);

/**
* @fn void track_highlight_time(track_t *track, double time)
 * @brief set highlight cursor to arbitrary time in track editor
 * @param track the track
 * @param time	time to display in editor
 * @ingroup track
 *
 * Set cursor to arbitrary time in track editor.
 * Previous highlighting is removed.
 */
FTS_API void track_highlight_time(track_t *track, double time);

/**
* @fn void track_highlight_events_and_time(track_t *track, double time, int n, event_t *event[])
 * @brief highlight events in track editor  and set cursor to given time
 * @param track the track
 * @param time	time to display in editor
 * @param n	number of events in event array
 * @param event	array of track events
 * @ingroup track
 *
 * Highlight events in track editor (up to 64) and set cursor to arbitrary time.
 * Set n to 0 and event to NULL to only display time.
 * Previous highlighting is removed.
 */
FTS_API void track_highlight_events_and_time(track_t *track, double time, int n, event_t *event[]);

/**
* @fn void track_highlight_cluster(track_t *track, event_t *event, event_t *next)
 * @brief highlight sequential event group in track editor
 * @param track the track
 * @param event pointer to first event
 * @param event pointer to last event
 * @ingroup track
 *
 * Highlight sequential event group in track editor.
 * Previous highlighting is removed.
 */
FTS_API void track_highlight_cluster(track_t *track, event_t *event, event_t *next);

/**
* @fn event_t *track_highlight_and_next(track_t *track, event_t *event)
 * @brief highlight event in track editor and get next event
 * @param track the track
 * @param event the event
 * @return next event in track
 * @ingroup track
 *
 * Highlight one event in track editor, set cursor on its starting time.
 * Previous highlighting is removed.
 */
FTS_API event_t *track_highlight_and_next(track_t *track, event_t *event);

/**
* @fn void track_highlight_reset(track_t *track)
 * @brief reset highlighted events in track
 * @param track	pointer to track
 * @ingroup track
 */
FTS_API void track_highlight_reset(track_t *track);

/** @}	end of doc group highlighting */

#endif  /* _SEQUENCE_TRACK_H_ */
